@@ -13,8 +13,10 @@ import time
 from datetime import datetime, timedelta
 from contextlib import redirect_stdout, redirect_stderr
 import code
+import random
 
 from pyspark.sql import SparkSession, SQLContext, Row
+from pyspark import SparkFiles
 
 {% if use_instance_principle %}
 USE_INSTANCE_PRINCIPLE = True
@@ -24,13 +26,71 @@ OCI_CONFIG = json.loads("""{{ oci_config_str }}""")
 OCI_KEY = """{{ oci_key }}"""
 {% endif %}
 
+random.seed()
+
+def get_server_channel(region, run_dir):
+    # from spark_etl.core import ServerChannelInterface
+    class ServerChannel:
+        def __init__(self, region, run_dir):
+            # run_dir point to the current job's run dir, not the base run dir
+            self.spark = None
+            self.region = region
+            self.run_dir = run_dir
+
+            o = urlparse(run_dir)
+            self.namespace = o.netloc.split('@')[1]
+            self.bucket = o.netloc.split('@')[0]
+            self.object_name_prefix = o.path[1:]  # drop the leading /
+
+
+        def read_json(self, spark, name):
+            from oci_core import os_download_json
+            os_client = get_os_client_ex(spark, self.region)
+            return os_download_json(
+                os_client,
+                self.namespace, self.bucket,
+                os.path.join(self.object_name_prefix, name)
+            )
+
+
+        def has_json(self, spark, name):
+            from oci_core import os_has_object
+            os_client = get_os_client_ex(spark, self.region)
+            return os_has_object(
+                os_client,
+                self.namespace, self.bucket,
+                os.path.join(self.object_name_prefix, name)
+            )
+
+
+        def write_json(self, spark, name, payload):
+            from oci_core import os_upload_json
+            os_client = get_os_client_ex(spark, self.region)
+            os_upload_json(
+                os_client,
+                payload,
+                self.namespace, self.bucket,
+                os.path.join(self.object_name_prefix, name)
+            )
+
+
+        def delete_json(self, spark, name):
+            from oci_core import os_delete_object
+            os_client = get_os_client_ex(spark, self.region)
+            os_delete_object(
+                os_client,
+                self.namespace, self.bucket,
+                os.path.join(self.object_name_prefix, name)
+            )
+    
+    return ServerChannel(region, run_dir)
 
 # lib installer
-def _install_libs(lib_url, run_id):
+def _install_libs(run_id):
     current_dir = os.getcwd()
     base_dir    = os.path.join(current_dir, run_id)
     lib_dir     = os.path.join(base_dir, 'python_libs')
-    lib_zip     = os.path.join(base_dir, 'lib.zip')
+    lib_zip     = SparkFiles.get("lib.zip")
     lock_name   = os.path.join(base_dir, '__lock__')
 
     os.makedirs(base_dir, exist_ok=True)
@@ -43,21 +103,20 @@ def _install_libs(lib_url, run_id):
                 if not os.path.isdir(lib_dir):
                     print("_install_libs: install lib starts")
                     os.makedirs(lib_dir)
-                    subprocess.check_call(['wget', "-O", lib_zip, lib_url])
-                    subprocess.check_call(['unzip', lib_zip, "-d", lib_dir])
+                    subprocess.check_call(['unzip', "-qq", lib_zip, "-d", lib_dir])
                     print("_install_libs: install lib done")
                 if lib_dir not in sys.path:
                     print(f"_install_libs: add {lib_dir} path")
                     sys.path.insert(0, lib_dir)
-                return 
+                return
             finally:
                 os.remove(lock_name)
         except OSError as e:
             if e.errno == errno.EEXIST:
-                time.sleep(10)
+                time.sleep(random.randint(1, 10))
                 continue
             raise
-    
+
     raise Exception("Failed to install libraries!")
 
 class Asker:
@@ -65,7 +124,7 @@ class Asker:
         self.spark = None
         self.run_dir = run_dir
         self.app_region = app_region
-    
+
     def initialize(self, spark):
         from oci_core import dfapp_get_os_client, get_delegation_token, get_os_client
 
@@ -95,7 +154,7 @@ class Asker:
         start_time = datetime.utcnow()
         while True:
             r = self.os_client.list_objects(
-                namespace, 
+                namespace,
                 bucket,
                 prefix = answer_name,
                 limit = 1
@@ -109,16 +168,6 @@ class Asker:
             return os_download_json(self.os_client, namespace, bucket, answer_name)
 
 
-def _save_result(os_client, run_dir, result):
-    from oci_core import os_upload_json
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    os_upload_json(os_client, result, namespace, bucket, f"{object_name_prefix}/result.json")
-
 def get_os_client_ex(spark, region):
     from oci_core import dfapp_get_os_client, get_delegation_token, get_os_client
     if USE_INSTANCE_PRINCIPLE:
@@ -133,71 +182,6 @@ def get_os_client_ex(spark, region):
     return os_client
 
 
-def _read_json(spark, region, run_dir, name):
-    from oci_core import os_download_json
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    os_client = get_os_client_ex(spark, region)
-    return os_download_json(os_client, namespace, bucket, f"{object_name_prefix}/{name}")
-
-def _has_json(spark, region, run_dir, name):
-    from oci_core import os_download_json
-    import oci
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    os_client = get_os_client_ex(spark, region)
-
-    try:
-        os_client.head_object(namespace, bucket, f"{object_name_prefix}/{name}")
-        return True
-    except oci.exceptions.ServiceError as e:
-        if e.status==404:
-            return False
-        raise
-
-def _write_json(spark, region, run_dir, name, payload):
-    from oci_core import os_upload_json
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    os_client = get_os_client_ex(spark, region)
-    os_upload_json(os_client, payload, namespace, bucket, f"{object_name_prefix}/{name}")
-
-def _delete_json(spark, region, run_dir, name):
-    from oci_core import os_upload_json
-    import oci
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    os_client = get_os_client_ex(spark, region)
-    os_client.delete_object(namespace, bucket, f"{object_name_prefix}/{name}")
-
-
-def _get_args(os_client, run_dir):
-    from oci_core import os_download_json
-    o = urlparse(run_dir)
-
-    namespace = o.netloc.split('@')[1]
-    bucket = o.netloc.split('@')[0]
-    object_name_prefix = o.path[1:]  # drop the leading /
-
-    return os_download_json(os_client, namespace, bucket, f"{object_name_prefix}/args.json")
-
-
 def _bootstrap():
     parser = argparse.ArgumentParser(description='job')
     parser.add_argument(
@@ -209,9 +193,6 @@ def _bootstrap():
     parser.add_argument(
         "--run-dir", type=str, required=True, help="Run Directory",
     )
-    parser.add_argument(
-        "--lib-url", type=str, required=True, help="Library URL",
-    )
     parser.add_argument(
         "--app-region", type=str, required=True, help="Application Region",
     )
@@ -225,138 +206,39 @@ def _bootstrap():
     spark = SparkSession.builder.appName("RunJob").getOrCreate()
     sc = spark.sparkContext
     sc.addPyFile(f"{args.deployment_location}/app.zip")
+    sc.addFile(f"{args.deployment_location}/lib.zip")
     print("Application loaded")
 
-    subprocess.check_call(['wget', args.lib_url])
-
     # The archive file goes into /opt/dataflow
     # Load python libraries
     current_dir = os.getcwd()
     lib_dir = os.path.join(current_dir, 'python_libs')
     os.mkdir(lib_dir)
+    lib_zip = SparkFiles.get("lib.zip")
     subprocess.call([
-        'unzip', "lib.zip", "-d", lib_dir
+        'unzip', "-qq", lib_zip, "-d", lib_dir
     ])
     sys.path.insert(0, lib_dir)
 
     run_id = args.run_id
     run_dir = args.run_dir
 
-    # The entry is in a python file "main.py"
-    from oci_core import dfapp_get_os_client, get_delegation_token, get_os_client
-    if USE_INSTANCE_PRINCIPLE:
-        delegation_token = get_delegation_token(spark)
-        os_client = dfapp_get_os_client(args.app_region, delegation_token)
-    else:
-        with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as key_f:
-            key_f.write(OCI_KEY)
-        _oci_config = dict(OCI_CONFIG)
-        _oci_config['key_file'] = key_f.name
-        os_client = get_os_client(None, config=_oci_config)
-    xargs = _get_args(os_client, run_dir)
+    server_channel = get_server_channel(args.app_region, run_dir)
+    xargs = server_channel.read_json(spark, "args.json")
+
     entry = importlib.import_module("main")
     result = entry.main(spark, xargs, sysops={
-        "install_libs": lambda : _install_libs(args.lib_url, run_id),
-        "read_json": lambda name: _read_json(spark, args.app_region, run_dir, name),
-        "write_json": lambda name, payload: _write_json(spark, args.app_region, run_dir, name, payload),
-        "delete_json": lambda name: _delete_json(spark, args.app_region, run_dir, name),
-        "has_json": lambda name: _has_json(spark, args.app_region, run_dir, name),
+        "install_libs": lambda : _install_libs(run_id),
         "ask": Asker(run_dir, args.app_region),
-        "cli_main": cli_main,
+        "channel": server_channel
     })
+    server_channel.write_json(spark, "result.json", result)
 
-    # user need to initialize ask with ask.initialize(spark) before using it
-
-    _save_result(os_client, run_dir, result)
-
-    # TODO: make a copy of the log file
 
 class PySparkConsole(code.InteractiveInterpreter):
     def __init__(self, locals=None):
         super(PySparkConsole, self).__init__(locals=locals)
 
-def handle_pwd(user_input, write_json):
-    write_json(
-        "cli-response.json", 
-        {
-            "status": "ok",
-            "output": os.getcwd()
-        }
-    )
-
-def handle_bash(user_input, write_json):
-    cmd_buffer = '\n'.join(user_input['lines'])
-    f = io.StringIO()
-    with redirect_stdout(f):
-        p = subprocess.run(cmd_buffer, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-    
-    write_json(
-        "cli-response.json", 
-        {
-            "status": "ok",
-            "exit_code": p.returncode,
-            "output": p.stdout.decode('utf-8'),
-        }
-    )
-
-def handle_python(user_input, console, write_json):
-    source = '\n'.join(user_input['lines'])
-    stdout_f = io.StringIO()
-    stderr_f = io.StringIO()
-    with redirect_stdout(stdout_f):
-        with redirect_stderr(stderr_f):
-            console.runsource(source, symbol="exec")
-    
-    write_json(
-        "cli-response.json", 
-        {
-            "status": "ok",
-            "output": stdout_f.getvalue() + "\n" + stderr_f.getvalue() ,
-        }
-    )
-
-def cli_main(spark, args, sysops={}):
-    read_json = sysops['read_json']
-    write_json = sysops['write_json']
-    delete_json = sysops['delete_json']
-    has_json = sysops['has_json']
-    console = PySparkConsole(locals={'spark': spark})
-
-    write_json(
-        "cli-response.json", 
-        {
-            "status": "ok",
-            "output": "Welcome to OCI Spark-CLI Interface",
-        }
-    )
-
-    while True:
-        if not has_json('cli-request.json'):
-            time.sleep(1)
-            continue
-
-        user_input = read_json('cli-request.json')
-        delete_json('cli-request.json')
-
-        if user_input["type"] == "@@quit":
-            write_json(
-                "cli-response.json", 
-                {
-                    "status": "ok",
-                    "output": "Server quit gracefully",
-                }
-            )
-            break
-        if user_input["type"] == "@@pwd":
-            handle_pwd(user_input, write_json)
-            continue
-        if user_input["type"] == "@@bash":
-            handle_bash(user_input, write_json)
-            continue
-        if user_input["type"] == "@@python":
-            handle_python(user_input, console, write_json)
-            continue
-    return {"status": "ok"}
 
 
 _bootstrap()
